import os
import csv
import glob
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
import scipy.stats as stats
from scipy.stats import exponweib
from scipy.stats import spearmanr
from scipy.optimize import fmin
import numpy as np
from sklearn.linear_model import LinearRegression
from statsmodels.stats.outliers_influence import variance_inflation_factor
import statsmodels.api as sm 
import re
from statistics import mean, median
import statistics

# replace the paths with the paths where the files are locally stored
PackageMetricsPath = "D:\\Package analysis study\\CSVs\\Package-all projects\\all_package_metrics_111_projects_final.csv"

#read the csv file into a dataframe
df = pd.read_csv(PackageMetricsPath) 

# classification of package sizes (large, moderate, and small)
per_25 = np.percentile(df['NumCls'], 25)
per_90 = np.percentile(df['NumCls'], 90)
print("Threshold for small package", per_25)
print("Threshold for large package", per_90)

# distribution plot
plt.style.use('seaborn')
sns.set(font_scale=1.1)
sns.distplot(df['NumCls'], axlabel="Number of classes", hist=True, kde=True, rug=True, bins=100, color='darkblue', hist_kws={'edgecolor':'black'}, kde_kws={'linewidth':1}, rug_kws={'color':'black'})
plt.tight_layout()
plt.axvline(x=22, color='r', linewidth=1.25)
plt.text(23,0.074,'90%', color='r')
plt.axvline(x=2, color='r', linewidth=1.25)
plt.text(-5,0.074,'25%', color='r')

# Cohen's f2 effect size method
def cohenf2(Rab, Ra):
    return (Rab-Ra)/(1-Rab) 

# identifying coupling issues in large packages        
def largePackagesWithCopuling(PackageMetricsPath, couplingMetric, good, bad, regular, listForLargePackagesWithCoupling):    
    with open(PackageMetricsPath, 'r') as read_package_file:
        package_reader = csv.DictReader(read_package_file) # read the file having package metrics
        for package_row in package_reader:
            if int(package_row['NumCls']) > 22 and int(package_row[couplingMetric]) > 16: # condition for large package and bad coupling
                bad = bad + 1
                listForLargePackagesWithCoupling.append(int(package_row[couplingMetric])) 
            if int(package_row['NumCls']) > 22 and int(package_row[couplingMetric]) <= 6: # condition for large package and good coupling
                good = good + 1 
                listForLargePackagesWithCoupling.append(int(package_row[couplingMetric])) 
            if int(package_row['NumCls']) > 22 and (int(package_row[couplingMetric]) > 6 and int(package_row[couplingMetric]) <= 16): # condition for large package and regular coupling
                regular = regular + 1
                listForLargePackagesWithCoupling.append(int(package_row[couplingMetric]))        
    print("Number of large packages with good", couplingMetric, good)
    print("Number of large packages with bad", couplingMetric, bad)
    print("Number of large packages with regular", couplingMetric, regular)
    print("Mean:", statistics.mean(listForLargePackagesWithCoupling))
    print("Median:", statistics.median(listForLargePackagesWithCoupling))
    print("Max:", max(listForLargePackagesWithCoupling))
    print("Min:", min(listForLargePackagesWithCoupling))
    print("SD:", statistics.stdev(listForLargePackagesWithCoupling))
    
    
# identifying coupling issues in small packages
def smallPackagesWithCoupling(PackageMetricsPath, couplingMetric, good, bad, regular, listForSmallPackagesWithCoupling):    
    with open(PackageMetricsPath, 'r') as read_package_file:
        package_reader = csv.DictReader(read_package_file) # read the file having package metrics
        for package_row in package_reader:
            if (int(package_row['NumCls']) == 1 or int(package_row['NumCls']) == 2) and int(package_row[couplingMetric]) > 16: # condition for small package and bad coupling
                bad = bad + 1
                listForSmallPackagesWithCoupling.append(int(package_row[couplingMetric]))
            if (int(package_row['NumCls']) == 1 or int(package_row['NumCls']) == 2) and int(package_row[couplingMetric]) <= 6: # condition for small package and good coupling
                good = good + 1
                listForSmallPackagesWithCoupling.append(int(package_row[couplingMetric]))
            if (int(package_row['NumCls']) == 1 or int(package_row['NumCls']) == 2) and (int(package_row[couplingMetric]) > 6 and int(package_row[couplingMetric]) <= 16): # condition for small package and regular coupling
                regular = regular + 1
                listForSmallPackagesWithCoupling.append(int(package_row[couplingMetric]))
    print("Number of small packages with good", couplingMetric, good)
    print("Number of small packages with bad", couplingMetric, bad)
    print("Number of small packages with regular", couplingMetric, regular) 
    print("Mean:", statistics.mean(listForSmallPackagesWithCoupling))
    print("Median:", statistics.median(listForSmallPackagesWithCoupling))
    print("Max:", max(listForSmallPackagesWithCoupling))
    print("Min:", min(listForSmallPackagesWithCoupling))
    
    
# identifying coupling issues in moderate packages
def moderatePackagesWithCoupling(PackageMetricsPath, couplingMetric, good, bad, regular, listForModeratePackagesWithCoupling):    
    with open(PackageMetricsPath, 'r') as read_package_file:
        package_reader = csv.DictReader(read_package_file) # read the file having package metrics
        for package_row in package_reader:
            if (int(package_row['NumCls']) > 2 and int(package_row['NumCls']) <= 22) and int(package_row[couplingMetric]) > 16: # condition for small package and bad coupling
                bad = bad + 1
                listForModeratePackagesWithCoupling.append(int(package_row[couplingMetric]))
            if (int(package_row['NumCls']) > 2 and int(package_row['NumCls']) <= 22) and int(package_row[couplingMetric]) <= 6: # condition for small package and good coupling
                good = good + 1
                listForModeratePackagesWithCoupling.append(int(package_row[couplingMetric]))
            if (int(package_row['NumCls']) > 2 and int(package_row['NumCls']) <= 22) and (int(package_row[couplingMetric]) > 6 and int(package_row[couplingMetric]) <= 16): # condition for small package and regular coupling
                regular = regular + 1
                listForModeratePackagesWithCoupling.append(int(package_row[couplingMetric]))
    print("Number of moderate packages with good", couplingMetric, good)
    print("Number of moderate packages with bad", couplingMetric, bad)
    print("Number of moderate packages with regular", couplingMetric, regular) 
    print("Mean:", statistics.mean(listForModeratePackagesWithCoupling))
    print("Median:", statistics.median(listForModeratePackagesWithCoupling))
    print("Max:", max(listForModeratePackagesWithCoupling))
    print("Min:", min(listForModeratePackagesWithCoupling))
    
# identifying cohesion issues in large, moderate and small packages
def packagesWithCohesion(PackageMetricsPath, cohesionMetric, listForLargePackagesWithCohesion, listForSmallPackagesWithCohesion, listForModeratePackagesWithCohesion):    
    with open(PackageMetricsPath, 'r') as read_package_file:
        package_reader = csv.DictReader(read_package_file) # read the file having package metrics
        for package_row in package_reader:
            if int(package_row['NumCls']) > 22:
                listForLargePackagesWithCohesion.append(int(package_row[cohesionMetric]))
            if (int(package_row['NumCls']) == 1 or int(package_row['NumCls']) == 2):
                listForSmallPackagesWithCohesion.append(int(package_row[cohesionMetric]))
            if (int(package_row['NumCls']) > 2 and int(package_row['NumCls']) <= 22):
                listForModeratePackagesWithCohesion.append(int(package_row[cohesionMetric]))
    print("Mean large:", statistics.mean(listForLargePackagesWithCohesion))
    print("Median large:", statistics.median(listForLargePackagesWithCohesion))
    print("Max large:", max(listForLargePackagesWithCohesion))
    print("Min large:", min(listForLargePackagesWithCohesion))
    print("Mean small:", statistics.mean(listForSmallPackagesWithCohesion))
    print("Median small:", statistics.median(listForSmallPackagesWithCohesion))
    print("Max small:", max(listForSmallPackagesWithCohesion))
    print("Min small:", min(listForSmallPackagesWithCohesion))
    print("Mean moderate:", statistics.mean(listForModeratePackagesWithCohesion))
    print("Median moderate:", statistics.median(listForModeratePackagesWithCohesion))
    print("Max moderate:", max(listForModeratePackagesWithCohesion))
    print("Min moderate:", min(listForModeratePackagesWithCohesion))

# identifying complexity issues in large, moderate, and small packages
def packagesWithComplexity(PackageMetricsPath, complexityMetric, listForLargePackagesWithComplexity, listForSmallPackagesWithComplexity, listForModeratePackagesWithComplexity):    
    with open(PackageMetricsPath, 'r') as read_package_file:
        package_reader = csv.DictReader(read_package_file) # read the file having package metrics
        for package_row in package_reader:
            if int(package_row['NumCls']) > 22:
                listForLargePackagesWithComplexity.append(int(package_row[complexityMetric]))
            if (int(package_row['NumCls']) == 1 or int(package_row['NumCls']) == 2):
                listForSmallPackagesWithComplexity.append(int(package_row[complexityMetric]))
            if (int(package_row['NumCls']) > 2 and int(package_row['NumCls']) <= 22):
                listForModeratePackagesWithComplexity.append(int(package_row[complexityMetric]))
    print("Mean large:", statistics.mean(listForLargePackagesWithComplexity))
    print("Median large:", statistics.median(listForLargePackagesWithComplexity))
    print("Max large:", max(listForLargePackagesWithComplexity))
    print("Min large:", min(listForLargePackagesWithComplexity))
    print("Mean small:", statistics.mean(listForSmallPackagesWithComplexity))
    print("Median small:", statistics.median(listForSmallPackagesWithComplexity))
    print("Max small:", max(listForSmallPackagesWithComplexity))
    print("Min small:", min(listForSmallPackagesWithComplexity))
    print("Mean moderate:", statistics.mean(listForModeratePackagesWithComplexity))
    print("Median moderate:", statistics.median(listForModeratePackagesWithComplexity))
    print("Max moderate:", max(listForModeratePackagesWithComplexity))
    print("Min moderate:", min(listForModeratePackagesWithComplexity))                       
            
# Calculating VIF
print("--------------------------")
print("Variance Inflation Factor")
print("--------------------------")
X = df[['Ce', 'Ca', 'ConnComp', 'H', 'R', 'NumOpsCls', 'NumCls_tc']] #independent and control variables
vif_data = pd.DataFrame() 
vif_data["Metric"] = X.columns
vif_data["VIF"] = [variance_inflation_factor(X.values, i)
                          for i in range(len(X.columns))] 
print(vif_data) # printing the VIF scores

print("------------------------------------------------")
print("Descriptive Statistics and Kruskal-Wallis H-test")
print("------------------------------------------------")
count = 0
good = 0
bad = 0
regular = 0
couplingMetric = 'Ce'
listForSmallPackagesWithCoupling = []
listForLargePackagesWithCoupling = []
listForModeratePackagesWithCoupling = []

print("Descriptive statistics of coupling")
largePackagesWithCopuling(PackageMetricsPath,couplingMetric, good, bad, regular, listForLargePackagesWithCoupling)
smallPackagesWithCoupling(PackageMetricsPath,couplingMetric, good, bad, regular, listForSmallPackagesWithCoupling)
moderatePackagesWithCoupling(PackageMetricsPath, couplingMetric, good, bad, regular, listForModeratePackagesWithCoupling)
# performing H-test to see the significant difference in large, moderate, and small packages in terms of coupling
print(stats.kruskal(listForLargePackagesWithCoupling, listForSmallPackagesWithCoupling, listForModeratePackagesWithCoupling))

cohesionMetric = 'ConnComp'
listForLargePackagesWithCohesion = []
listForSmallPackagesWithCohesion = []
listForModeratePackagesWithCohesion = []

print("Descriptive statistics of cohesion")
packagesWithCohesion(PackageMetricsPath, cohesionMetric, listForLargePackagesWithCohesion, listForSmallPackagesWithCohesion, listForModeratePackagesWithCohesion)
# performing H-test to see the significant difference in large, moderate, and small packages in terms of cohesion
print(stats.kruskal(listForLargePackagesWithCohesion, listForSmallPackagesWithCohesion, listForModeratePackagesWithCohesion))

complexityMetric = 'R'
listForLargePackagesWithComplexity = []
listForSmallPackagesWithComplexity = []
listForModeratePackagesWithComplexity = []

print("Descriptive statistics of complexity")
packagesWithComplexity(PackageMetricsPath, complexityMetric, listForLargePackagesWithComplexity, listForSmallPackagesWithComplexity, listForModeratePackagesWithComplexity)
# performing H-test to see the significant difference in large, moderate, and small packages in terms of complexity
print(stats.kruskal(listForLargePackagesWithComplexity, listForSmallPackagesWithComplexity, listForModeratePackagesWithComplexity))

# calculating risk ratio of the coupling metrics in terms of good, regular, and bad
print("--------------------------------------------------")
print("Risk Ratio of large packages w.r.t small packages")
print("--------------------------------------------------")
table = sm.stats.Table2x2(np.array([[258,437],[2057,185]]))
print("Risk ratio of good Ca \n", table.summary())

table = sm.stats.Table2x2(np.array([[224,471],[33,2209]]))
print("Risk ratio of bad Ca \n", table.summary())

table = sm.stats.Table2x2(np.array([[213,482],[152,2090]]))
print("Risk ratio of regular Ca \n", table.summary())

table = sm.stats.Table2x2(np.array([[123,572],[2066,176]]))
print("Risk ratio of good Ce \n", table.summary())

table = sm.stats.Table2x2(np.array([[381,314],[28,2214]]))
print("Risk ratio of bad Ce \n", table.summary())

table = sm.stats.Table2x2(np.array([[191,504],[148,2094]]))
print("Risk ratio of regular Ce \n", table.summary())

print("----------------------------------------------------")
print("Risk Ratio of large packages w.r.t moderate packages")
print("----------------------------------------------------")
table = sm.stats.Table2x2(np.array([[258,437],[3134,914]]))
print("Risk ratio of good Ca \n", table.summary())

table = sm.stats.Table2x2(np.array([[224,471],[246,3802]]))
print("Risk ratio of bad Ca \n", table.summary())

table = sm.stats.Table2x2(np.array([[213,482],[668,3380]]))
print("Risk ratio of regular Ca \n", table.summary())

table = sm.stats.Table2x2(np.array([[123,572],[2489,1559]]))
print("Risk ratio of good Ce \n", table.summary())

table = sm.stats.Table2x2(np.array([[381,314],[464,3584]]))
print("Risk ratio of bad Ce \n", table.summary())

table = sm.stats.Table2x2(np.array([[191,504],[1095,2953]]))
print("Risk ratio of regular Ce \n", table.summary())

# Regression analysis
print("--------------------")
print("Regression Analysis")
print("--------------------")
X = df[['Ce', 'Ca', 'ConnComp', 'H', 'R', 'NumOpsCls', 'NumCls_tc']] # indepdendent variables
Y = df['NumCls'] # dependent variable
X2 = sm.add_constant(X)
est = sm.OLS(Y, X2)
reg = est.fit()
print(reg.summary())

# calculating cohen's f2 of the maintainability metrics
print("-----------------------")
print("Cohen's f2 Effect Size")
print("-----------------------")
print('Effect size (Ce): %.3f' % cohenf2(0.814,0.814))
print('Effect size (Ca): %.3f' % cohenf2(0.814,0.814))
print('Effect size (ConnComp): %.3f' % cohenf2(0.814,0.768))
print('Effect size (H): %.3f' % cohenf2(0.814,0.808))
print('Effect size (R): %.3f' % cohenf2(0.814,0.776))
# Effect sizes of the control variables
print('Effect size (NumOpsCls): %.3f' % cohenf2(0.814,0.757))
print('Effect size (NumCls_tc): %.3f' % cohenf2(0.814,0.813))

# computing Spearman's correlations of package size with other maintainability metrics
print("-----------------------")
print("Spearman's Correlation")
print("-----------------------")
print("Correlation between NumCls and Ce", spearmanr(df['NumCls'],df['Ce']))

print("Correlation between NumCls and Ca", spearmanr(df['NumCls'],df['Ca']))

print("Correlation between NumCls and Connected components", spearmanr(df['NumCls'],df['ConnComp']))

print("Correlation between NumCls and Relational cohesion", spearmanr(df['NumCls'],df['H']))

print("Correlation between NumCls and R", spearmanr(df['NumCls'],df['R']))
